#ifndef CMAP_H
#define CMAP_H

//#define CONVERT_MAP 1

#include <QImage>
#include <QColor>
#include "system/cmaths.h"
#include <vector>

#define TILE_SIZE   32

enum EWallSide
{
    eWallSideUp = 0,
    eWallSideDown,
    eWallSideLeft,
    eWallSideRight,
    eWallSideMax
};

#define eWallSideMonster (EWallSide)(eWallSideMax + 1)

enum EParamType
{
    eParamOrnate = 0,
    eParamDoor,
    eParamDoorOrnate,
    eParamEnum,
    eParamInt,
    eParamBool,
    eParamScript,
    eParamStack,
    eParamString
};

//---------------------------------------------------------------------------------------------
struct CParamType
{
    EParamType  mType;
    QString     mName;
    QStringList mValues;
};

//---------------------------------------------------------------------------------------------
class CParam
{
public:
    EParamType  mType;
};

class CParamOrnate : public CParam
{
public:
    CParamOrnate();
    uint8_t mValue;
};

class CParamInt : public CParam
{
public:
    CParamInt();
    int32_t mValue;
};

class CParamBool : public CParam
{
public:
    CParamBool();
    bool mValue;
};

class CParamDoor : public CParam
{
public:
    CParamDoor();
    uint8_t mValue;
};

class CParamDoorOrnate : public CParam
{
public:
    CParamDoorOrnate();
    uint8_t mValue;
};

class CParamEnum : public CParam
{
public:
    CParamEnum();
    uint8_t mValue;
};

class CParamScript : public CParam
{
public:
    CParamScript();
    std::string mValue;
};

class CParamStack : public CParam
{
public:
    CParamStack();
};

class CParamString : public CParam
{
public:
    CParamString();
    std::string mValue;
};

//---------------------------------------------------------------------------------------------
class CWall
{
public:
    CWall();
    virtual ~CWall();

    void    setType(uint8_t type);
    uint8_t getType();
    CWall&  operator=(const CWall& rhs);
    void    load(FILE* handle);
    void    save(FILE* handle);

    CParam*     findParamByName(QString name);
    uint8_t     getOrnateParam(QString name);
    int32_t     getIntParam(QString name);
    bool        getBoolParam(QString name);
    void        setBoolParam(QString name, bool value);
    uint8_t     getEnumParam(QString name);
    void        setEnumParam(QString name, uint8_t value);
    std::string getStringParam(QString name);
    void        setStringParam(QString name, std::string value);
    std::string getScriptParam(QString name);

    std::vector<CParam*>    mParams;

private:
    void    deleteParams();

    uint8_t mType;
};

//---------------------------------------------------------------------------------------------
class CTile
{
public:
    CTile();
    virtual ~CTile();

    void    setType(uint8_t type);
    uint8_t getType();
    void    copyWithoutWalls(const CTile& rhs);
    CTile&  operator=(const CTile& rhs);
    void    load(FILE* handle);
    void    save(FILE* handle);

    CParam*     findParamByName(QString name);
    uint8_t     getOrnateParam(QString name);
    uint8_t     getDoorParam(QString name);
    uint8_t     getDoorOrnateParam(QString name);
    uint8_t     getEnumParam(QString name);
    int32_t     getIntParam(QString name);
    void        setIntParam(QString name, int32_t value);
    bool        getBoolParam(QString name);
    void        setBoolParam(QString name, bool value);
    std::string getScriptParam(QString name);

    std::vector<CParam*>    mParams;
    CWall   mWalls[eWallSideMax];

private:
    void    deleteParams();

    uint8_t mType;
};

//---------------------------------------------------------------------------------------------
class CObject
{
public:
    CObject();
    CObject(const CObject& rhs);
    virtual ~CObject();

    void        setType(uint8_t type);
    uint8_t     getType();
    CObject&    operator=(const CObject& rhs);
    void        load(FILE* handle);
    void        save(FILE* handle);

    CParam*     findParamByName(QString name);
    std::string getStringParam(QString name);
    void        setStringParam(QString name, std::string value);
    int32_t     getIntParam(QString name);
    void        setIntParam(QString name, int32_t value);

    std::vector<CParam*>    mParams;

private:
    void    deleteParams();

    uint8_t mType;
};

//---------------------------------------------------------------------------------------------
class CObjectStack
{
public:
    CObjectStack();
    CObjectStack(const CObjectStack& rhs);
    virtual ~CObjectStack();

    void            addObject(CObject object);
    CObject&        getObject(int index);
    void            removeObject(int index);
    size_t          getSize();
    CObjectStack&   operator=(const CObjectStack& rhs);
    void            load(FILE* handle);
    void            save(FILE* handle);

    CVec2       mPos;
    EWallSide   mSide;

private:
    std::vector<CObject>    mObjects;
};

//---------------------------------------------------------------------------------------------
class CMonsterGroup
{
public:
    CMonsterGroup();
    CMonsterGroup(const CMonsterGroup& rhs);
    virtual ~CMonsterGroup();

    void    setType(uint8_t type);
    uint8_t getType();
    CMonsterGroup&  operator=(const CMonsterGroup& rhs);
    void    load(FILE* handle);
    void    save(FILE* handle);

    CParam*     findParamByName(QString name);
    int32_t     getIntParam(QString name);
    uint8_t     getEnumParam(QString name);

    CVec2       mPos;
    std::vector<CParam*>    mParams;

private:
    void    deleteParams();

    uint8_t mType;
};

//---------------------------------------------------------------------------------------------
class CMap
{
public:
    CMap();
    CMap(const CMap& rhs);
    virtual ~CMap();

    void    freeMemory();
    void    resize();

    void    setTile(CVec2 pos, uint8_t type);
    void    setWall(CVec2 pos, EWallSide side, uint8_t type);

    void    save(FILE* handle);
    void    load(FILE* handle);
    void    save(char* fileName);
    void    load(char* fileName);
    bool    isLoaded();

    CTile*  getTile(CVec2 pos);

    CObjectStack*   addObjectsStack(CVec2 pos, EWallSide side);
    CObjectStack*   findObjectsStack(CVec2 pos, EWallSide side);
    void            removeObjectsStack(CVec2 pos, EWallSide side);
    int             findMonsterGroupIndex(CVec2 pos);
    CMonsterGroup*  addMonsterGroup(CVec2 pos);
    CMonsterGroup*  findMonsterGroup(CVec2 pos);
    void            removeMonsterGroup(CVec2 pos);

    CMap&  operator=(const CMap& rhs);

    CVec2   mSize;
    static std::vector<std::vector<CParamType>>  mTilesParams; // list of the params for each type of tile
    static std::vector<std::vector<CParamType>>  mWallsParams; // list of the params for each type of wall
    static std::vector<std::vector<CParamType>>  mObjectsParams; // list of the params for each type of object
    static std::vector<std::vector<CParamType>>  mMonstersParams; // list of the params for each type of monster
    std::vector<CObjectStack>   mObjectsStacks;
    std::vector<CMonsterGroup>  mMonsterGroups;

private:
    int     findObjectsStackIndex(CVec2 pos, EWallSide side);

    CTile*  mTiles;
};

extern CMap  map;

#endif // CMAP_H
